# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import copy
# choosing multiprocessing over threading for clean Control-C interrupts (provides terminate())
from urlparse import urlparse
from multiprocessing import Process, Manager
from vcstools.vcs_base import VcsError

class MultiProjectException(Exception): pass

def conditional_abspath(uri):
  """
  @param uri: The uri to check
  @return: abspath(uri) if local path otherwise pass through uri
  """
  u = urlparse(uri)
  if u.scheme == '': # maybe it's a local file?
    return os.path.abspath(uri)
  else:
    return uri

def is_web_uri(source_uri):
  if source_uri is None or source_uri == '':
    return False
  parsed_uri = urlparse(source_uri)
  if (parsed_uri.scheme == ''
      and parsed_uri.netloc == ''
      and not '@' in parsed_uri.path.split('/')[0]):
    return False
  return True

def normabspath(localname, path):
  """
  if localname is absolute, return it normalized. If relative, return normalized join of path and localname
  """
  if os.path.isabs(localname) or path is None:
    return os.path.realpath(localname)
  abs_path = os.path.realpath(os.path.join(path, localname))
  return abs_path

def realpath_relation(abspath1, abspath2):
  """
  Computes the relationship abspath1 to abspath2
  :returns: None, 'SAME_AS', 'PARENT_OF', 'CHILD_OF'
  """
  assert os.path.isabs(abspath1)
  assert os.path.isabs(abspath2)
  realpath1 = os.path.realpath(abspath1)
  realpath2 = os.path.realpath(abspath2)
  if os.path.dirname(realpath1) == os.path.dirname(realpath2):
    if os.path.basename(realpath1) == os.path.basename(realpath2):
      return 'SAME_AS'
    return None
  else:
    commonprefix = os.path.commonprefix([realpath1, realpath2])
    if commonprefix == realpath1:
      return 'PARENT_OF'
    elif commonprefix == realpath2:
      return 'CHILD_OF'
  return None

def select_element(elements, localname):
  """
  selects entry among elements where path or localname matches.
  Prefers localname matches in case of ambiguity.
  """
  path_candidate = None
  if localname is not None:
    realpath = os.path.realpath(localname)
    for element in elements:
      if localname == element.get_local_name():
        path_candidate = element
        break
      elif realpath == os.path.realpath(element.get_path()):
        path_candidate = element
  return path_candidate


def select_elements(config, localnames):
  """
  selects config elements with given localnames, returns in the order given in config
  If localnames has one element which is path of the config, return all elements
  """
  if config is None:
    return []
  if localnames is None:
    return config.get_config_elements()
  elements = config.get_config_elements()
  selected = []
  notfound = []
  for localname in localnames:
    element = select_element(elements, localname)
    if element is not None:
      selected.append(element)
    else:
      notfound.append(localname)
  if notfound != []:
    raise MultiProjectException("Unknown elements '%s'"%notfound)
  result = []
  # select in order and remove duplicates
  for element in config.get_config_elements():
    if element in selected:
      result.append(element)
  if result == []:
      if (len(localnames) == 1
          and os.path.realpath(localnames[0]) == os.path.realpath(config.get_base_path())):
        return config.get_elements()
  return result
  
  
## Multithreading The following classes help with distributing work
## over several instances, providing wrapping for starting, joining,
## collecting results, and catching Exceptions. Also they provide
## support for running groups of threads sequentially, for the case
## that some library is not thread-safe.
  
class WorkerThread(Process):

  def __init__(self, worker, outlist, index):
    Process.__init__(self)
    self.worker = worker
    if worker is None or worker.element is None:
      raise MultiProjectException("Bug: Invalid Worker")
    self.outlist = outlist
    self.index = index

  def run(self):
    result = {}
    try:
      result = {'entry': self.worker.element.get_path_spec()}
      result_dict = self.worker.do_work()
      if result_dict is not None:
        result.update(result_dict)
      else:
        result.update({'error': MultiProjectException("worker returned None")})
    except MultiProjectException as e:
      result.update({'error': e})
    except VcsError as e:
      result.update({'error': e})
    except OSError as e:
      result.update({'error': e})
    except Exception as e:
      # this would be a bug, and we need trace to find them in multithreaded cases.
      import sys, traceback
      traceback.print_exc(file=sys.stderr)
      result.update({'error': e})
    self.outlist[self.index] = result

class DistributedWork():
  
  def __init__(self, capacity, num_threads=10, silent=True):
    man = Manager() # need managed array since we need the results later
    self.outputs = man.list([None for _ in range(capacity)])
    self.threads = []
    self.sequentializers = {}
    self.index = 0
    self.num_threads = num_threads
    self.silent = silent
    
  def add_thread(self, worker):
    thread = WorkerThread(worker, self.outputs, self.index)
    if self.index >= len(self.outputs):
      raise MultiProjectException("Bug: Declared capacity exceeded %s >= %s"%(self.index, len(self.outputs)))
    self.index += 1
    self.threads.append(thread)
    
  # def add_to_sequential_thread_group(self, worker, group):
  #   """Workers in each sequential thread group run one after the other, groups run in parallel"""
  #   class ThreadSequentializer(Process):
  #     """helper class to run 'threads' one after the other"""
  #     def __init__(self):
  #       Process.__init__(self)
  #       self.workers = []
  #     def add_worker(self, worker):
  #       self.workers.append(worker)
  #     def run(self):
  #     for worker in self.workers:
  #       worker.run() # not calling start on purpose
  #   thread = WorkerThread(worker, self.outputs, self.index)
  #   if self.index >= len(self.outputs):
  #     raise MultiProjectException("Bug: Declared capacity exceeded %s >= %s"%(self.index, len(self.outputs)))
  #   self.index += 1
  #   if group not in self.sequentializers:
  #     self.sequentializers[group] = ThreadSequentializer()
  #     self.sequentializers[group].add_worker(thread)
  #     self.threads.append(self.sequentializers[group])
  #   else:
  #     self.sequentializers[group].add_worker(thread)
    
  def run(self):
    """
    Execute all collected workers, terminate all on KeyboardInterrupt
    """
    if self.threads == []:
      return []
    if (self.num_threads == 1):
      for thread in self.threads:
        thread.run()
    else:
      # The following code is rather delicate and may behave differently
      # using threading or multiprocessing. running_threads is
      # intentionally not used as a shrinking list because of al the
      # possible multithreading / interruption corner cases
      # Not using Pool because of KeyboardInterrupt cases
      try:
        waiting_index = 0
        maxthreads = self.num_threads
        running_threads = []
        missing_threads = copy.copy(self.threads)
        while len(running_threads) > 0 or len(missing_threads) > 0:
          if len(running_threads) < maxthreads:
            to_index = min(waiting_index + maxthreads - len(running_threads), len(self.threads))
            for i in range(waiting_index, to_index):
              self.threads[i].start()
            waiting_index = to_index
          running_threads = [t for t in self.threads if t is not None and t.is_alive()]
          if (not self.silent
              and len(running_threads) > 0):
            print("[%s] still active"%",".join([th.worker.element.get_local_name() for th in running_threads]))
          for thread in running_threads:
            if thread in missing_threads:
              missing_threads.remove(thread)
            thread.join(1)
      except KeyboardInterrupt as k:
        for thread in self.threads:
          if thread is not None and thread.is_alive():
            print("[%s] terminated while active"%thread.worker.element.get_local_name())
            thread.terminate()
        raise k
    
    self.outputs = filter(lambda x: x is not None, self.outputs)
    message = ''
    for output in self.outputs:
      if "error" in output:
        if 'entry' in output:
          message += "Error processing '%s' : %s\n"%(output['entry'].get_local_name(), output["error"])
        else:
          message += "%s\n"%output["error"]
    if message != '':
      raise MultiProjectException(message)
    return self.outputs



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os

class ROSInstallException(Exception): pass

ROSINSTALL_FILENAME = ".rosinstall"


def is_path_stack(path):
  """
  
  @return: True if the path provided is the root of a stack.
  """
  stack_path = os.path.join(path, 'stack.xml')
  if os.path.isfile(stack_path):
    return True
  return False

def is_path_ros(path):
  """
  warning: exits with code 1 if stack document is invalid
  @param path: path of directory to check
  @type  path: str
  @return: True if path points to the ROS stack
  @rtype: bool
  """
  if path is None:
    return False
  stack_path = os.path.join(path, 'stack.xml')
  if os.path.isfile(stack_path):
    return 'ros' == os.path.basename(path)
  return False


def get_ros_stack_path(config):
  rp = None
  for t in config.get_config_elements():
    if is_path_ros(t.get_path()):
      if rp is not None:
        raise ROSInstallException("Two ros roots defined in config, delete one: %s %s"%(rp, t.get_path()))
      rp = t.get_path()
  return rp


def get_ros_package_path(config):
  """ Return the simplifed ROS_PACKAGE_PATH """
  code_trees = []
  for t in reversed(config.get_config_elements()):
    if not is_path_ros(t.get_path()):
      if not os.path.isfile(t.get_path()):
        code_trees.append(t.get_path())
  return code_trees




# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

import os

from helpers import ROSInstallException, get_ros_stack_path, get_ros_package_path, ROSINSTALL_FILENAME
from config_elements import SetupConfigElement

CATKIN_CMAKE_TOPLEVEL="""#
#  TOPLEVEL cmakelists
#
cmake_minimum_required(VERSION 2.8)
cmake_policy(SET CMP0003 NEW)
cmake_policy(SET CMP0011 NEW)

set(CMAKE_CXX_FLAGS_INIT "-Wall")

enable_testing()

include(${CMAKE_SOURCE_DIR}/workspace-config.cmake OPTIONAL)

list(APPEND CMAKE_PREFIX_PATH ${CMAKE_BINARY_DIR} ${CMAKE_BINARY_DIR}/cmake)

file(MAKE_DIRECTORY ${CMAKE_BINARY_DIR}/lib)
file(MAKE_DIRECTORY ${CMAKE_BINARY_DIR}/bin)

if (IS_DIRECTORY ${CMAKE_SOURCE_DIR}/catkin)
  message(STATUS "+++ catkin")
  set(CATKIN_BUILD_PROJECTS "ALL" CACHE STRING
    "List of projects to build, or ALL for all.  Use to completely exclude certain projects from cmake traversal.")
  add_subdirectory(catkin)
else()
  find_package(catkin)
endif()

catkin_workspace()
"""

def generate_catkin_cmake(path, catkinpp):
    with open(os.path.join(path, "CMakeLists.txt"), 'w') as cmake_file:
      cmake_file.write(CATKIN_CMAKE_TOPLEVEL)

    if catkinpp:
      with open(os.path.join(path, "workspace-config.cmake"), 'w') as config_file:
        config_file.write("set (CMAKE_PREFIX_PATH %s)"%catkinpp)

def generate_setup_sh_text(workspacepath):
  # overlay or standard
  text =  """#!/usr/bin/env sh
# THIS IS A FILE AUTO-GENERATED BY rosinstall
# IT IS UNLIKELY YOU WANT TO EDIT THIS FILE BY HAND
# IF YOU WANT TO CHANGE THE ROS ENVIRONMENT VARIABLES
# USE THE rosinstall OR rosws TOOL INSTEAD.
# see: http://www.ros.org/wiki/rosinstall
"""
# Sadly we cannot infer the workspacepath from within the sourced file
  text += """\nexport ROS_WORKSPACE=%s
if [ ! "$ROS_MASTER_URI" ] ; then export ROS_MASTER_URI=http://localhost:11311 ; fi
unset ROS_ROOT
""" % workspacepath
   
# use python script to read ros_package_path and setup-file elements
  text += """
# python script to read .rosinstall even when rosnistall is not installed
export _PARSED_CONFIG=`/usr/bin/env python << EOPYTHON
import sys, os, yaml;
filename = '.rosinstall'
if 'ROS_WORKSPACE' in os.environ:
  filename = os.path.join(os.environ['ROS_WORKSPACE'], filename)
if not os.path.isfile(filename):
    sys.exit("There is no file at %s"%filename)
with open(filename, "r") as f:
  try:
    v=f.read();
  except Exception as e:
    sys.exit("Failed to read file: %s %s "%(filename, str(e)))
try:
  y = yaml.load(v);
except Exception as e:
  sys.exit("Invalid yaml in %s: %s "%(filename, str(e)))
if y is not None:
  lnames=[x.values()[0]['local-name'] for x in y if x.values() is not None and x.keys()[0] != "setup-file"]
  paths = [os.path.normpath(os.path.join(os.environ['ROS_WORKSPACE'], z)) for z in lnames if not os.path.isfile(os.path.join(os.environ['ROS_WORKSPACE'], z))]
  output=''
  if len(paths) > 0:
    output += ':'.join(reversed(paths))
  output += 'ROSINSTALL_PATH_SETUPFILE_SEPARATOR'
  snames=[x.values()[0]['local-name'] for x in y if x.values() is not None and x.keys()[0] == "setup-file"]
  paths = [os.path.join(os.environ['ROS_WORKSPACE'], z) for z in snames if os.path.isfile(os.path.join(os.environ['ROS_WORKSPACE'], z))]
  output += ':'.join(paths)
  print(output)
EOPYTHON`

#whitespace separates results
_ROS_PACKAGE_PATH_ROSINSTALL=`echo "$_PARSED_CONFIG" | sed -s 's,\(.*\)ROSINSTALL_PATH_SETUPFILE_SEPARATOR\(.*\),\\1,'`
_SETUPFILES_ROSINSTALL=`echo "$_PARSED_CONFIG" | sed -s 's,\(.*\)'ROSINSTALL_PATH_SETUPFILE_SEPARATOR'\(.*\),\\2,'`
unset _PARSED_CONFIG

# colon separates entries
_LOOP_SETUP_FILE=`echo $_SETUPFILES_ROSINSTALL | sed -s 's,\([^:]*\)[:]\(.*\),\\1,'`
while [ ! -z "$_LOOP_SETUP_FILE" ]
do
  if [ -f "$_LOOP_SETUP_FILE" ]; then
    . $_LOOP_SETUP_FILE
  else
    echo warn: no such file : "$_LOOP_SETUP_FILE"
  fi
  _SETUPFILES_ROSINSTALL=`echo $_SETUPFILES_ROSINSTALL | sed -s 's,\([^:]*[:]*\),,'`
  _LOOP_SETUP_FILE=`echo $_SETUPFILES_ROSINSTALL | sed -s 's,\([^:]*\)[:]\(.*\),\\1,'`
done

unset _LOOP_SETUP_FILE
unset _SETUPFILES_ROSINSTALL

export ROS_PACKAGE_PATH=$_ROS_PACKAGE_PATH_ROSINSTALL
unset _ROS_PACKAGE_PATH_ROSINSTALL

# using ROS_ROOT now being in ROS_PACKAGE_PATH
export _ROS_ROOT_ROSINSTALL=`/usr/bin/env python << EOPYTHON
import sys, os;
if 'ROS_PACKAGE_PATH' in os.environ:
  pkg_path = os.environ['ROS_PACKAGE_PATH']
  for path in pkg_path.split(':'):
    if (os.path.basename(path) == 'ros'
        and os.path.isfile(os.path.join(path, 'stack.xml'))):
      print(path)
      break
EOPYTHON`
if [ ! -z "${_ROS_ROOT_ROSINSTALL}" ]; then
  export ROS_ROOT=$_ROS_ROOT_ROSINSTALL
  export PATH=$ROS_ROOT/bin:$PATH
  export PYTHONPATH=$ROS_ROOT/core/roslib/src:$PYTHONPATH
fi
unset _ROS_ROOT_ROSINSTALL
"""

  return text

def generate_setup_bash_text(shell, no_ros = False):
  if shell == 'bash':
    script_path = """
SCRIPT_PATH="${BASH_SOURCE[0]}";
if([ -h "${SCRIPT_PATH}" ]) then
  while([ -h "${SCRIPT_PATH}" ]) do SCRIPT_PATH=`readlink "${SCRIPT_PATH}"`; done
fi
export OLDPWDBAK=$OLDPWD
pushd . > /dev/null
cd `dirname ${SCRIPT_PATH}` > /dev/null
SCRIPT_PATH=`pwd`;
popd  > /dev/null
export OLDPWD=$OLDPWDBAK
"""
  elif shell == 'zsh':
    script_path = 'SCRIPT_PATH="$(dirname $0)"';
  else:
    raise ROSInstallException("%s shell unsupported."%shell);

  text =  """#!/usr/bin/env %(shell)s
# THIS IS A FILE AUTO-GENERATED BY rosinstall
# IT IS UNLIKELY YOU WANT TO EDIT THIS FILE BY HAND
# IF YOU WANT TO CHANGE THE ROS ENVIRONMENT VARIABLES
# USE THE rosinstall TOOL INSTEAD.
# see: http://www.ros.org/wiki/rosinstall

CATKIN_SHELL=%(shell)s

%(script_path)s

# Load the path of this particular setup.%(shell)s

if [ ! -f "$SCRIPT_PATH/setup.sh" ]; then
  echo "Bug: shell script unable to determine its own location: $SCRIPT_PATH"
  return 22
fi


# unset _ros_decode_path to check later whether setup.sh has sourced ros%(shell)s
unset -f _ros_decode_path 1> /dev/null 2>&1

. $SCRIPT_PATH/setup.sh"""%locals()

  if not no_ros:
    text += """
# Cannot rely on $? due to set -o errexit in build scripts
RETURNCODE=`type _ros_decode_path 2> /dev/null | grep function 1>/dev/null 2>&1 || echo error`

if [ ! "$RETURNCODE" = "" ]; then
  RETURNCODE=`rospack help 1> /dev/null 2>&1 || echo error`
  if  [ "$RETURNCODE" = "" ]; then
    ROSSHELL_PATH=`rospack find rosbash`/rosbash
    if [ -e "$ROSSHELL_PATH" ]; then
      . $ROSSHELL_PATH
    fi
  else
    echo "rospack could not be found, you cannot have ros%(shell)s features until you bootstrap ros"
  fi
fi
"""%locals()
  return text


def generate_setup(config, no_ros_allowed = False):
  ros_root = get_ros_stack_path(config)
  if not ros_root:
    if not no_ros_allowed:
      raise ROSInstallException("""
No 'ros' stack detected in candidates %s.
Please add the location of a ros distribution to this command.

See http://ros.org/wiki/rosinstall."""%([t.get_path() for t in config.get_config_elements() if os.path.basename(t.get_local_name())=='ros']) )
  
  text = generate_setup_sh_text(workspacepath = config.get_base_path())
  setup_path = os.path.join(config.get_base_path(), 'setup.sh')
  with open(setup_path, 'w') as f:
    f.write(text)


  for shell in ['bash', 'zsh']:
    text = generate_setup_bash_text(shell, ros_root is None)
    setup_path = os.path.join(config.get_base_path(), 'setup.%s'%shell)
    with open(setup_path, 'w') as f:
      f.write(text)





# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import os
import re
from optparse import OptionParser
from common import MultiProjectException

# Support for any command line interface (CLI) for rosinstall

def get_workspace(argv, shell_path, config_filename = None, varname = None):
  """
  If target option -t is given return value of that one. Else, varname
  is given and exists, considers that one, plus,
  if config_filename is given, searches for a file named in config_filename
  in '.', above '.'.
  In that case, if two solutions are found, asks the user.
  
  :returns: abspath if a .rosinstall was found, error and exist else.
  """
  parser = OptionParser()
  parser.add_option("-t", "--target-workspace", dest="workspace", default=None,
                      help="which workspace to use",
                      action="store")
  # suppress errors based on invalid option
  argv2 = filter(lambda x: (not x.startswith('-') or x.startswith('--target-workspace=') or x.startswith('-t=') or x in ['-t', '--target-workspace']), argv)
  (options, args) = parser.parse_args(argv2)
  if options.workspace is not None:
    if (config_filename is not None
        and not os.path.isfile(os.path.join(options.workspace, config_filename))):
      raise MultiProjectException("%s has no workspace configuration file '%s'"%(os.path.abspath(options.workspace), config_filename))
    return os.path.abspath(options.workspace)

  varname_path = None
  if varname is not None and varname in os.environ:
    # workspace could be relative, maybe confusing, but that's the users fault
    varname_path = os.environ[varname]
    if varname_path.strip() == '' or not os.path.isdir(varname_path):
      varname_path = None

  # use current dir
  current_path = None
  if config_filename is not None:
    while shell_path is not None and not shell_path == os.path.dirname(shell_path):
      if os.path.exists(os.path.join(shell_path, config_filename)):
        current_path = shell_path
        break
      shell_path = os.path.dirname(shell_path)

  if current_path is not None and varname_path is not None and not os.path.samefile(current_path, varname_path):
    raise MultiProjectException("Ambiguous workspace: %s=%s, %s"%(varname, varname_path, os.path.abspath(config_filename)))

  if current_path is None and varname_path is None:
    raise MultiProjectException("Command requires a target workspace.")

  if current_path is not None:
    return current_path
  else:
    return varname_path

def _uris_match(basepath, uri1, uri2):
  """
  True if uri2 is None or not None and same folder or equal string
  as uri1. Relative folders resolved using basepath
  """
  if uri1 is None:
    uri1 = ''
  if uri2 is None:
    return True
  if (uri1 == uri2
      or (basepath is not None
          and os.path.isdir(os.path.join(basepath, uri1))
          and os.path.realpath(os.path.join(basepath, uri2)) == os.path.realpath(os.path.join(basepath, uri1)))):
      return True
  return False
      
  
def _get_status_flags(basepath, line):
  if 'exists' in line and line['exists'] is False:
    return 'x'
  mflag = ''
  if 'modified' in line and line['modified'] is True:
    mflag = 'M'
  if (('curr_uri' in line
       and not _uris_match(basepath, line['uri'], line['curr_uri']))
      or ('specversion' in line
          and line['specversion'] is not None
          and line['actualversion'] is not None
          and line['specversion'] != line['actualversion'])):
    mflag += 'V'
  return mflag

def get_info_table(basepath, entries, data_only = False, reverse = False):
  headers = {
    'uri':"URI  (Spec) (https://...)",
    'scm':"SCM ",
    'localname':"Localname",
    'version':"Version-Spec",
    'matching': "UID  (Spec)",
    'status':"S"
    }

  # table design
  selected_headers=['localname', 'status', 'scm', 'version', 'matching', 'uri'  ]

  column_length = {}

  for header in headers.keys():
    column_length[header] = len(headers[header])

  outputs = []
  for line in entries:
    if not 'curr_uri' in line:
      line['curr_uri'] = None
    if not 'specversion' in line:
      line['specversion'] = None
    if not 'actualversion' in line:
      line['actualversion'] = None
    if not 'curr_version' in line:
      line['curr_version'] = None
    if not 'version' in line:
      line['version'] = None
    output_dict = {'scm': line['scm'], 'uri': line['uri'], 'curr_uri': None, 'version': line['version'], 'localname': line['localname']}

    if line == None:
      print("Bug Warning, an element is missing")
      continue
    if line['scm'] == 'git':
      if (line['specversion'] is not None and len(line['specversion']) > 12):
        line['specversion'] = line['specversion'][0:12]
      if (line['actualversion'] is not None and len(line['actualversion'])>12):
        line['actualversion'] = line['actualversion'][0:12]
    if line['scm'] != None:
      if line['scm'] == 'svn':
        # in case of SVN, we can use the final part of standard uri as version
        uri = line['uri']
        version = line['version']
        m = re.match('(.*/)((tags|branches|trunk)(/.*)*)', uri)
        if m is not None and len(m.groups()) > 1 and uri == ''.join(m.groups()[0:2]):
          uri = m.groups()[0]
          if (m.groups()[1] is not None and version is None or version.strip() == ''):
            version = m.groups()[1]
          else:
            version = m.groups()[1]
          line['uri'] = uri
          line['version'] = version
        if line['curr_uri'] is not None:
          uri = line['curr_uri']
          m = re.match('(.*/)((tags|branches|trunk)(/.*)*)', uri)
          if m is not None and len(m.groups()) > 1 and uri == ''.join(m.groups()[0:2]):
            uri = m.groups()[0]
            if (m.groups()[1] is not None and version is None or version.strip() == ''):
              version = m.groups()[1]
            else:
              version = m.groups()[1]
            line['curr_uri'] = uri
            line['curr_version'] = version
      if (line['curr_version'] is not None
          and line['version'] != line['curr_version']):
        output_dict['version'] = "%s  (%s)"%(line['curr_version'], line['version'])
      else:
        output_dict['version'] = line['version']
      if (line['specversion'] is not None
        and line['specversion'] != ''
        and line['actualversion'] != line['specversion']):
        output_dict['matching'] = "%s (%s)"%( line['actualversion'], line['specversion'])
      else:
        output_dict['matching'] = line['actualversion']
      common_prefix = "https://"
      if line['uri'].startswith(common_prefix):
        line['uri'] = line['uri'][len(common_prefix):]
        output_dict['uri'] = line['uri']
      if (line['curr_uri'] is not None
          and line['curr_uri'].startswith(common_prefix)):
        line['curr_uri'] = line['curr_uri'][len(common_prefix):]
      if (not _uris_match(basepath, line['uri'], line['curr_uri'])):
        output_dict['uri'] = "%s  (%s)"%(line['curr_uri'], line['uri'])
    else:
      output_dict['matching'] = " "
    output_dict['status'] = _get_status_flags(basepath, line)
        
    for header in headers.keys():
      if output_dict[header] != None:
        column_length[header] = max(column_length[header], len(output_dict[header]))
    outputs.append(output_dict)

  resultlines = []
  if not data_only and len(outputs) > 0:
    header_line = ' '
    for i, header in enumerate(selected_headers):
      output = headers[header]
      if i < len(selected_headers) -1:
        output = output.ljust(column_length[header]) + " "
      header_line += output
    resultlines.append(header_line)
    header_line=' '
    for i, header in enumerate(selected_headers):
      output = '-' * len(headers[header])
      if i < len(selected_headers) -1:
        output = output.ljust(column_length[header]) + " "
      header_line += output
    resultlines.append(header_line)

  if reverse:
    outputs = reversed(outputs)
  for entry in outputs:
    if entry == None:
      print("Bug Warning, an element is missing")
      continue
    data_line = ' '
    for i, header in enumerate(selected_headers):
      output = entry[header]
      if output == None:
        output = ''
      if i < len(selected_headers) -1:
        output = output.ljust(column_length[header]) + " "
      data_line += output
    resultlines.append(data_line)
  return "\n".join(resultlines)


def get_info_list(basepath, line, data_only = False):
  """
  Info for a single config entry
  """

  if line == None:
    print("Bug Warning, an element is missing")
    return

  headers = {
    'uri'           : "URI:",
    'curr_uri'      : "Current URI:",
    'scm'           : "SCM:",
    'localname'     : "Localname:",
    'path'          : "Path",
    'version'       : "Version-Spec:",
    'status'        : "Status:",
    'specversion'   : "Spec-Revision:",
    'actualversion' : "Current-Revision:",
    'properties'    : "Other Properties:"
    }

  # table design
  selected_headers=['localname', 'path', 'status', 'scm', 'uri', 'curr_uri', 'version', 'specversion', 'actualversion', 'properties' ]

  line['status'] = _get_status_flags(basepath, line)

  header_length = 0
  for header in headers.keys():
    header_length = max(header_length, len(headers[header]))
  result = ''
  for header in selected_headers:
    if not data_only:
      title = "%s  "%headers[header].ljust(header_length)
    else:
      title = ''
    if header in line:
      output = line[header]
    if output == None:
      output = ''
    result += "%s\n"%"%s%s"%(title, output)
  return result
